# -*- coding: utf-8 -*-

from config import Config

# Pretty Print table in tabular format
def pretty_print(table, justify = "L"):
    # get width for every column
    column_widths = [0] * table[0].__len__()
    offset = 3
    for row in table:
        for index, col in enumerate(row):
            width = len(str(col).decode("utf-8"))
            if width > column_widths[index]:
                column_widths[index] = width
    outputStr = ""
    for row_index, row in enumerate(table):
        rowList = []
        for col_index, col in enumerate(row):
            if justify == "R": # justify right
                formated_column = str(col).decode("utf-8").rjust(column_widths[col_index] + offset)
            elif justify == "L": # justify left
                formated_column = str(col).decode("utf-8").ljust(column_widths[col_index] + offset)
            elif justify == "C": # justify center
                formated_column = str(col).decode("utf-8").center(column_widths[col_index] + offset)
            rowList.append(formated_column.encode("utf-8"))
        if row_index == table.__len__()-1:
            outputStr += ' '.join(rowList)
        else:
            outputStr += ' '.join(rowList) + "\n"
    return outputStr

def get_new_contact_template(addressbook_name):
    return """# new contact
# Address book: %s
# if you want to cancel, exit without saving

# first and last name
# at least enter first name, first and last name or an organisation
First name   = 
Last name    = 
Organisation = 

# phone numbers
# format: PhoneX = type: number
# allowed types:
#   Standard: cell, home, work
#   Alternatively you can use every custom label (only letters). But maybe not all address book
#   clients support that.
Phone1 = cell: 
Phone2 = Dresden: 

# email addresses
# format: EmailX = type: address
# allowed types:
#   Standard: home, work
#   or a custom label (only letters)
Email1 = home: 

# post addresses
# format: AddressX = type: street and house number; postcode; city; region; country
# the region is optional so the following is allowed too:
# format: AddressX = type: street and house number; postcode; city;; country
# allowed types:
#   Standard: home, work
#   or a custom label (only letters)
Address1 = home: ; ; ; ; %s

# instant messaging and social networks
Jabber  = 
Skype   = 
Twitter = 
Webpage = 

# Miscellaneous stuff
# Birthday: day.month.year
Birthday = 
Nickname = """ % (addressbook_name, Config().get_default_country())

def get_existing_contact_template(vcard):
    strings = []
    for line in get_new_contact_template(vcard.get_addressbook_name()).splitlines():
        if line == "# new contact":
            strings.append("# Edit contact: %s" % vcard.get_full_name())
        elif line.lower().startswith("# if you want to cancel"):
            continue
        elif line.lower().startswith("first name"):
            strings.append("First name   = %s" % vcard.get_first_name())
        elif line.lower().startswith("last name"):
            strings.append("Last name    = %s" % vcard.get_last_name())
        elif line.lower().startswith("organisation"):
            strings.append("Organisation = %s" % vcard.get_organisation())
        elif line.lower().startswith("phone"):
            if line.lower().startswith("phone1"):
                if vcard.get_phone_numbers().__len__() == 0:
                    strings.append("Phone1 = cell: ")
                else:
                    for index, entry in enumerate(vcard.get_phone_numbers()):
                        strings.append("Phone%d = %s: %s" % (index+1, entry['type'], entry['value']))
        elif line.lower().startswith("email"):
            if line.lower().startswith("email1"):
                if vcard.get_email_addresses().__len__() == 0:
                    strings.append("Email1 = home: ")
                else:
                    for index, entry in enumerate(vcard.get_email_addresses()):
                        strings.append("Email%d = %s: %s" % (index+1, entry['type'], entry['value']))
        elif line.lower().startswith("address"):
            if line.lower().startswith("address1"):
                if vcard.get_post_addresses().__len__() == 0:
                    strings.append("Address1 = home: ; ; ; ; %s" % Config().get_default_country())
                else:
                    for index, entry in enumerate(vcard.get_post_addresses()):
                        strings.append("Address%d = %s: %s; %s; %s; %s; %s" % (index+1, entry['type'],
                                entry['street_and_house_number'], entry['postcode'], entry['city'],
                                entry['region'], entry['country']))
        elif line.lower().startswith("jabber"):
            strings.append("Jabber  = %s" % vcard.get_jabber_id())
        elif line.lower().startswith("skype"):
            strings.append("Skype   = %s" % vcard.get_skype_id())
        elif line.lower().startswith("twitter"):
            strings.append("Twitter = %s" % vcard.get_twitter_id())
        elif line.lower().startswith("webpage"):
            strings.append("Webpage = %s" % vcard.get_webpage())
        elif line.lower().startswith("birthday") and vcard.get_birthday() != None:
            date = vcard.get_birthday()
            strings.append("Birthday = %.2d.%.2d.%.4d" % (date.day, date.month, date.year))
        elif line.lower().startswith("nickname"):
            strings.append("Nickname = %s" % vcard.get_nickname())
        else:
            strings.append(line)
    return '\n'.join(strings)




#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import tempfile, subprocess, os, sys, re
import argparse
import helpers
from email.header import decode_header
from config import Config
from carddav_object import CarddavObject
from version import khard_version


def create_new_contact(addressbook):
    # create temp file
    tf = tempfile.NamedTemporaryFile(mode='w+t', delete=False)
    temp_file_name = tf.name
    tf.write(helpers.get_new_contact_template(addressbook['name']))
    tf.close()

    # start vim to edit contact template
    child = subprocess.Popen([Config().get_editor(), temp_file_name])
    streamdata = child.communicate()[0]

    # read temp file contents after editing
    tf = open(temp_file_name, "r")
    new_contact_template = tf.read()
    tf.close()
    os.remove(temp_file_name)

    # create carddav object from temp file
    vcard = CarddavObject(addressbook['name'], addressbook['path'])
    vcard.process_user_input(new_contact_template)
    vcard.write_to_file()
    print "Creation successful\n\n%s" % vcard.print_vcard()


def modify_existing_contact(vcard):
    # get content template for contact
    old_contact_template = helpers.get_existing_contact_template(vcard)

    # create temp file and open it with the specified text editor
    tf = tempfile.NamedTemporaryFile(mode='w+t', delete=False)
    temp_file_name = tf.name
    tf.write(old_contact_template)
    tf.close()

    # start editor to edit contact template
    child = subprocess.Popen([Config().get_editor(), temp_file_name])
    streamdata = child.communicate()[0]

    # read temp file contents after editing
    tf = open(temp_file_name, "r")
    new_contact_template = tf.read()
    tf.close()
    os.remove(temp_file_name)

    # check if the user changed anything
    if old_contact_template != new_contact_template:
        vcard.process_user_input(new_contact_template)
        vcard.write_to_file(overwrite=True)
        print "Creation successful\n\n%s" % vcard.print_vcard()
    else:
        print "Nothing changed."


def merge_existing_contacts(first_vcard, second_vcard):
    # create temp files for each vcard

    # first
    old_first_contact_template = helpers.get_existing_contact_template(first_vcard)
    first_tf = tempfile.NamedTemporaryFile(mode='w+t', delete=False)
    first_temp_file_name = first_tf.name
    first_tf.write(old_first_contact_template)
    first_tf.close()

    # second
    old_second_contact_template = helpers.get_existing_contact_template(second_vcard)
    second_tf = tempfile.NamedTemporaryFile(mode='w+t', delete=False)
    second_temp_file_name = second_tf.name
    second_tf.write(old_second_contact_template)
    second_tf.close()

    # start editor to edit contact template
    child = subprocess.Popen([Config().get_merge_editor(), first_temp_file_name, second_temp_file_name])
    streamdata = child.communicate()[0]

    # first
    first_tf = open(first_temp_file_name, "r")
    new_first_contact_template = first_tf.read()
    first_tf.close()
    os.remove(first_temp_file_name)

    # second
    second_tf = open(second_temp_file_name, "r")
    new_second_contact_template = second_tf.read()
    second_tf.close()
    os.remove(second_temp_file_name)

    # make sure at least one of them changed
    if new_first_contact_template == old_first_contact_template:
        print("Merge unsuccessfull, First contact unchanged")
        return
    if new_second_contact_template != old_second_contact_template:
        print("Merge unsuccessfull, please only modify the first contact");
        return

    first_vcard.process_user_input(new_first_contact_template)
    while True:
        input_string = raw_input("Merging contacts %s and %s\n\nMerged\n\n%s\n\nTo be removed\n\n%s\n\nAre you sure? (y/n): " \
                % (first_vcard.get_full_name(), second_vcard.get_full_name(),
                    first_vcard.print_vcard(), second_vcard.print_vcard()))
        if input_string.lower() in ["", "n", "q"]:
            print "Canceled"
            sys.exit(0)
        if input_string.lower() == "y":
            break

    first_vcard.write_to_file(overwrite=True)
    second_vcard.delete_vcard_file()
    print "Merge successful\n\n%s" % first_vcard.print_vcard()


def list_contacts(selected_addressbooks, vcard_list):
    if selected_addressbooks.__len__() == 1:
        print "Address book: %s" % selected_addressbooks[0]
        table = [["Id", "Name", "Phone", "E-Mail"]]
    else:
        print "Address books: %s" % ', '.join(selected_addressbooks)
        table = [["Id", "Name", "Phone", "E-Mail", "Address book"]]
    for index, vcard in enumerate(vcard_list):
        row = []
        row.append(index+1)
        if vcard.get_nickname() != "" \
                and Config().show_nicknames():
            row.append("%s (Nickname: %s)" % (vcard.get_full_name(), vcard.get_nickname()))
        else:
            row.append(vcard.get_full_name())
        if vcard.get_phone_numbers().__len__() > 0:
            phone1 = vcard.get_phone_numbers()[0]
            row.append("%s: %s" % (phone1['type'], phone1['value']))
        else:
            row.append("")
        if vcard.get_email_addresses().__len__() > 0:
            email1 = vcard.get_email_addresses()[0]
            row.append("%s: %s" % (email1['type'], email1['value']))
        else:
            row.append("")
        if selected_addressbooks.__len__() > 1:
            row.append(vcard.get_addressbook_name())
        table.append(row)
    print helpers.pretty_print(table)


def choose_vcard_from_list(selected_addressbooks, vcard_list):
    if vcard_list.__len__() == 0:
        return None
    elif vcard_list.__len__() == 1:
        return vcard_list[0]
    else:
        list_contacts(selected_addressbooks, vcard_list)
        while True:
            input_string = raw_input("Enter Id: ")
            if input_string in ["", "q", "Q"]:
                sys.exit(0)
            try:
                vcard_id = int(input_string)
                if vcard_id > 0 and vcard_id <= vcard_list.__len__():
                    break
            except ValueError as e:
                pass
            print "Please enter an Id between 1 and %d or nothing or q to exit." % vcard_list.__len__()
        print ""
        return vcard_list[vcard_id-1]
 

def main():
    # create the args parser
    parser = argparse.ArgumentParser(description="Khard is a carddav address book for the console")
    parser.add_argument("-a", "--addressbook", default="",
            help="Specify address book names as comma separated list")
    parser.add_argument("-r", "--reverse", action="store_true", help="Sort contacts in reverse order")
    parser.add_argument("-s", "--search", default="",
            help="Search for contacts. Special format for merge command: \"Contact 1|Contact 2\"")
    parser.add_argument("-t", "--sort", default="alphabetical", 
            help="Sort contacts list. Possible values: alphabetical, addressbook")
    parser.add_argument("-v", "--version", action="store_true", help="Get current program version")
    parser.add_argument("action", nargs="?", default="",
            help="Possible actions: list, details, mutt, alot, phone, new, add-email, modify, merge, remove and source")
    args = parser.parse_args()

    # version
    if args.version == True:
        print "Khard version %s" % khard_version
        sys.exit(0)

    # validate value for action
    if args.action == "":
        args.action = Config().get_default_action()
    if args.action not in Config().get_list_of_actions():
        print "Unsupported action. Possible values are: %s" % ', '.join(Config().get_list_of_actions())
        sys.exit(1)

    # load address books which are defined in the configuration file
    addressbooks = Config().get_all_addressbooks()
    if args.addressbook == "":
        args.addressbook = ','.join(addressbooks.keys())

    # given address book name
    selected_addressbooks = []
    for name in args.addressbook.split(","):
        if Config().has_addressbook(name) == False:
            print "Error: The entered address book \"%s\" does not exist. Possible values are: %s" \
                    % (name, ', '.join(addressbooks.keys()))
            sys.exit(1)
        selected_addressbooks.append(name)

    # search parameter
    # may either contain one search term for a standard search or two terms, devided by a "|" to
    # search for two contacts to merge them
    search_terms = args.search.split("|")
    if len(search_terms) == 1:
        search_terms.append("")

    # sort criteria
    if args.sort not in ["alphabetical", "addressbook"]:
        print "Unsupported sort criteria. Possible values: alphabetical, addressbook"
        sys.exit(1)

    # create new contact
    if args.action == "new":
        if selected_addressbooks.__len__() != 1:
            print "Error: You must specify an address book for the new contact\n" \
                    "Possible values are: %s" % ', '.join(addressbooks.keys())
            sys.exit(1)
        create_new_contact(addressbooks[selected_addressbooks[0]])
        sys.exit(0)

    # add email address to contact or create a new one if necessary
    if args.action == "add-email":
        email_address = ""
        name = ""
        for line in sys.stdin:
            if line.startswith("From:"):
                try:
                    name = line[6:line.index("<")-1]
                    email_address = line[line.index("<")+1:line.index(">")]
                except ValueError as e:
                    email_address = line[6:].strip()
                break
        # reopen stdin to get user input
        sys.stdin = open('/dev/tty')
        print "Khard: Add email address to contact"
        if not email_address:
            print "Found no email address"
            sys.exit(1)
        print "Email address: %s" % email_address
        if not name:
            name = raw_input("Contact's name: ")
        else:
            # fix encoding of senders name
            name, encoding = decode_header(name)[0]
            if encoding:
                name = name.decode(encoding).encode("utf-8")
            # remove quotes from name string
            name = name.replace("\"", "")
            # query user input.
            user_input = raw_input("Contact's name [%s]: " % name)
            # if empty, use the extracted name from above
            name = user_input or name
        # search for an existing contact
        selected_vcard = choose_vcard_from_list(selected_addressbooks,
                Config().get_vcard_objects(selected_addressbooks, args.sort, args.reverse, name, True))
        if selected_vcard == None:
            # create new contact
            while True:
                input_string = raw_input("Contact %s does not exist. Do you want to create it (y/n)? " % name)
                if input_string.lower() in ["", "n", "q"]:
                    print "Canceled"
                    sys.exit(0)
                if input_string.lower() == "y":
                    break
            print "Available address books: %s" % ', '.join(selected_addressbooks)
            while True:
                book_name = raw_input("Address book [%s]: " % selected_addressbooks[0]) or selected_addressbooks[0]
                if Config().has_addressbook(book_name):
                    break
            selected_vcard = CarddavObject(addressbooks[book_name]['name'], addressbooks[book_name]['path'])
            while True:
                first_name = raw_input("First name: ")
                last_name = raw_input("Last name: ")
                organization = raw_input("Organization: ")
                if not first_name and not last_name and not organization:
                    print "Error: All fields are empty."
                else:
                    break
            selected_vcard.set_name_and_organisation(first_name.decode("utf-8"),
                    last_name.decode("utf-8"), organization.decode("utf-8"))
        # check if the contact already contains the email address
        for email_entry in selected_vcard.get_email_addresses():
            if email_entry['value'] == email_address:
                print "The contact %s already contains the email address %s" % (selected_vcard, email_address)
                sys.exit(0)
        # ask for the email label
        print "\nAdding email address %s to contact %s" % (email_address, selected_vcard)
        label = raw_input("email label [home]: ") or "home"
        # add email address to vcard object
        selected_vcard.add_email_address(
                label.decode("utf-8"), email_address.decode("utf-8"))
        # save to disk
        selected_vcard.write_to_file(overwrite=True)
        print "Done.\n\n%s" % selected_vcard.print_vcard()
        sys.exit(0)

    # create a list of all found vcard objects
    vcard_list = Config().get_vcard_objects(selected_addressbooks, args.sort, args.reverse, search_terms[0], False)

    # print phone application  friendly contacts table
    if args.action == "phone":
        phone_list = []
        regexp = re.compile(search_terms[0].replace(" ", ".*"), re.IGNORECASE)
        for vcard in vcard_list:
            for tel_entry in vcard.get_phone_numbers():
                phone_number_line = "%s\t%s\t%s" \
                        % (tel_entry['value'], vcard.get_full_name(), tel_entry['type'])
                if len(re.sub("\D", "", search_terms[0])) >= 3:
                    # the user likely searches for a phone number cause the search string contains
                    # at least three digits
                    # so we remove all non-digit chars from the phone number field and match against that
                    if regexp.search(re.sub("\D", "", tel_entry['value'])) != None:
                        phone_list.append(phone_number_line)
                else:
                    # the user doesn't search for a phone number so we can perform a standard search
                    # without removing all non-digit chars from the phone number string
                    if regexp.search(phone_number_line) != None:
                        phone_list.append(phone_number_line)
        print '\n'.join(phone_list)
        if len(phone_list) == 0:
            sys.exit(1)
        sys.exit(0)

    # print mutt friendly contacts table
    if args.action == "mutt":
        address_list = ["searching for '%s' ..." % search_terms[0]]
        regexp = re.compile(search_terms[0].replace(" ", ".*"), re.IGNORECASE)
        for vcard in vcard_list:
            for email_entry in vcard.get_email_addresses():
                email_address_line = "%s\t%s\t%s" \
                        % (email_entry['value'], vcard.get_full_name(), email_entry['type'])
                if regexp.search(email_address_line) != None:
                    address_list.append(email_address_line)
        print '\n'.join(address_list)
        if len(address_list) <= 1:
            sys.exit(1)
        sys.exit(0)

    # print alot friendly contacts table
    if args.action == "alot":
        address_list = []
        regexp = re.compile(search_terms[0].replace(" ", ".*"), re.IGNORECASE)
        for vcard in vcard_list:
            for email_entry in vcard.get_email_addresses():
                email_address_line = "\"%s %s\" <%s>" \
                        % (vcard.get_full_name(), email_entry['type'], email_entry['value'])
                if regexp.search(email_address_line) != None:
                    address_list.append(email_address_line)
        print '\n'.join(address_list)
        if len(address_list) == 0:
            sys.exit(1)
        sys.exit(0)

    # cancel if we found no contacts
    if vcard_list.__len__() == 0:
        print "Found no contacts"
        sys.exit(0)

    # print user friendly contacts table
    if args.action == "list":
        list_contacts(selected_addressbooks, vcard_list)
        sys.exit(0)

    # show source or details, modify or delete contact
    if args.action in ["details", "modify", "remove", "source"]:
        selected_vcard = choose_vcard_from_list(selected_addressbooks, vcard_list)

        if args.action == "details":
            print selected_vcard.print_vcard()

        elif args.action == "modify":
            modify_existing_contact(selected_vcard)

        elif args.action == "remove":
            while True:
                input_string = raw_input("Deleting contact %s. Are you sure? (y/n): " \
                        % selected_vcard.get_full_name())
                if input_string.lower() in ["", "n", "q"]:
                    print "Canceled"
                    sys.exit(0)
                if input_string.lower() == "y":
                    break
            selected_vcard.delete_vcard_file()
            print "Contact deleted successfully"

        elif args.action == "source":
            child = subprocess.Popen([Config().get_editor(),
                    selected_vcard.get_vcard_full_filename()])
            streamdata = child.communicate()[0]

    # merge two contacts
    if args.action == "merge":
        # get the first vcard, into which to merge
        # respect the users search query if available
        first_selected_vcard = choose_vcard_from_list(selected_addressbooks,
                Config().get_vcard_objects(selected_addressbooks, args.sort, args.reverse, search_terms[0], False))
        if first_selected_vcard == None:
            print("Found no contact to merge into")
            sys.exit(1)
        print("Merge into %s\n" % first_selected_vcard.get_full_name())

        # then get the second vcard, which to merge from
        # clear out the users potential serch query
        print("Now choose the vcard from which to merge:")
        second_selected_vcard = choose_vcard_from_list(selected_addressbooks,
                Config().get_vcard_objects(selected_addressbooks, args.sort, args.reverse, search_terms[1], False))
        if second_selected_vcard == None:
            print "Found no contact to merge from"
            sys.exit(1)

        # compare contacts
        if first_selected_vcard == second_selected_vcard:
            print("Selected same contact twice.")
            sys.exit(1)
        else:
            merge_existing_contacts(first_selected_vcard, second_selected_vcard)
            sys.exit(0)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# singleton code comes from:
# http://code.activestate.com/recipes/52558/#as_content

import os, sys, glob, re, operator
from configobj import ConfigObj
from carddav_object import CarddavObject

class Config:
    """ A python singleton """

    class __impl:
        """ Implementation of the singleton interface """
        def __init__(self):
            # load config file
            xdg_config_home = os.environ.get("XDG_CONFIG_HOME") or \
                    os.path.join(os.path.expanduser("~"), ".config")
            config_file = os.environ.get("KHARD_CONFIG") or \
                    os.path.join(xdg_config_home, "khard", "khard.conf")
            if os.path.exists(config_file) == False:
                print "Config file %s not available" % config_file
                sys.exit(2)
            self.config = ConfigObj(config_file, interpolation=False)

            # general settings
            if self.config.has_key("general") == False:
                print "Error in config file\nMissing main section \"[general]\"."
                sys.exit(2)
            if self.config['general'].has_key("editor") == False:
                print "Error in config file\nMissing editor parameter. Example: editor = /usr/bin/vim."
                sys.exit(2)
            elif os.path.exists(self.config['general']['editor']) == False:
                print "Error in config file\nInvalid editor path."
                sys.exit(2)
            if self.config['general'].has_key("merge_editor") == False:
                print "Error in config file\nMissing merge_editor parameter. Example: merge_editor = /usr/bin/vimdiff."
                sys.exit(2)
            elif os.path.exists(self.config['general']['merge_editor']) == False:
                print "Error in config file\nInvalid merge_editor path."
                sys.exit(2)
            if self.config['general'].has_key("default_country") == False:
                print "Error in config file\nMissing default country parameter."
                sys.exit(2)
            if self.config['general'].has_key("default_action") == False:
                print "Error in config file\nMissing default action parameter."
                sys.exit(2)
            elif self.config['general']['default_action'] not in self.get_list_of_actions():
                print "Error in config file\n" \
                        "Non existing value for default action parameter\n" \
                        "Possible values are: %s" % ', '.join(self.get_list_of_actions())
                sys.exit(2)
            if self.config['general'].has_key("show_nicknames") == False:
                self.config['general']['show_nicknames'] = False
            elif self.config['general']['show_nicknames'] == "yes":
                self.config['general']['show_nicknames'] = True
            elif self.config['general']['show_nicknames'] == "no":
                self.config['general']['show_nicknames'] = False
            else:
                print "Error in config file\nshow_nicknames parameter must be yes or no."
                sys.exit(2)

            # load address books
            if self.config.has_key("addressbooks") == False:
                print "Error in config file\nMissing main section \"[addressbooks]\"."
                sys.exit(2)
            if self.config['addressbooks'].keys().__len__() == 0:
                print "Error in config file\nNo address book entries available."
                sys.exit(2)
            for name in self.config['addressbooks'].keys():
                addressbook = self.config['addressbooks'][name]
                if addressbook.has_key("path") == False:
                    print "Error in config file\nMissing path to the \"%s\" address book." % name
                    sys.exit(2)
                if addressbook['path'].startswith("~"):
                    addressbook['path'] = addressbook['path'].replace("~", os.path.expanduser("~"))
                if os.path.exists(addressbook['path']) == False:
                    print "Error in config file\nThe path %s to the address book %s does not exist." % (addressbook['path'], name)
                    sys.exit(2)
                # set address book name
                addressbook['name'] = name
                # load all vcard files
                error_counter = 0
                addressbook['vcards'] = []
                for filename in glob.glob(os.path.join(addressbook['path'], "*.vcf")):
                    try:
                        addressbook['vcards'].append(
                                CarddavObject(addressbook['name'], addressbook['path'], filename))
                    except CarddavObject.VCardParseError as e:
                        error_counter += 1
                        print "Parse Error: %s" % e
                if error_counter == 1:
                    print "1 vcard file could not be parsed"
                elif error_counter > 1:
                    print "%d vcard files could not be parsed" % error_counter

        def get_editor(self):
            return self.config['general']['editor']

        def get_merge_editor(self):
            return self.config['general']['merge_editor']

        def get_default_country(self):
            return self.config['general']['default_country']

        def get_list_of_actions(self):
            return ["list", "details", "new", "add-email", "modify", "merge",
                    "remove", "mutt", "phone", "alot", "source"]

        def get_default_action(self):
            return self.config['general']['default_action']

        def show_nicknames(self):
            return self.config['general']['show_nicknames']

        def has_addressbook(self, name):
            return self.config['addressbooks'].has_key(name)

        def get_all_addressbooks(self):
            return self.config['addressbooks']

        def get_addressbook(self, name):
            try:
                return self.config['addressbooks'][name]
            except KeyError as e:
                print "The address book \"%s\" does not exist" % name
                sys.exit(3)

        def get_vcard_objects(self, addressbook_names, sort_criteria, reverse, search, strict_search):
            """returns a list of vcard objects
            :param addressbook_names: list of selected address books
            :type addressbook_names: list(str)
            :param sort_criteria: sort list by given criteria
            :type sort_criteria: str
            :param reverse: reverse ordering
            :type reverse: bool
            :param search: filter contact list
            :type search: str
            :param strict_search: if True, search only in full name field
            :type strict_search: bool
            :returns: list of vcard objects
            :rtype: list(vobject.vCard)
            """
            vcard_list = []
            regexp = re.compile(search.replace(" ", ".*"), re.IGNORECASE | re.DOTALL)
            for addressbook_name in addressbook_names:
                addressbook = self.get_addressbook(addressbook_name)
                for vcard in addressbook['vcards']:
                    if strict_search:
                        if regexp.search(vcard.get_full_name()) != None:
                            vcard_list.append(vcard)
                    else:
                        if regexp.search(vcard.print_vcard()) != None:
                            vcard_list.append(vcard)
                        else:
                            # special case for phone numbers without a space between prefix and number
                            for phone_entry in vcard.get_phone_numbers():
                                if regexp.search(re.sub("\D", "", phone_entry['value'])) != None:
                                    vcard_list.append(vcard)
                                    break
            if sort_criteria == "addressbook":
                return sorted(vcard_list,
                        key = lambda x: (x.get_addressbook_name().lower(),
                            x.get_full_name().lower()),
                        reverse=reverse)
            else:
                return sorted(vcard_list, key = lambda x: x.get_full_name().lower(), reverse=reverse)

    ####################################
    # storage for the instance reference
    ####################################
    __instance = None

    def __init__(self):
        """ Create singleton instance """
        # Check whether we already have an instance
        if Config.__instance is None:
            # Create and remember instance
            Config.__instance = Config.__impl()

        # Store instance reference as the only member in the handle
        self.__dict__['_Config__instance'] = Config.__instance

    def __getattr__(self, attr):
        """ Delegate access to implementation """
        return getattr(self.__instance, attr)

    def __setattr__(self, attr, value):
        """ Delegate access to implementation """
        return setattr(self.__instance, attr, value)




# -*- coding: utf-8 -*-

import os, sys, string, random, datetime, re
import vobject

class CarddavObject:
    def __init__(self, addressbook_name, addressbook_path, filename=""):
        self.addressbook_name = addressbook_name
        if filename == "":
            # create new vcard
            self.vcard = vobject.vCard()
            choice = string.ascii_uppercase + string.digits
            uid_obj = self.vcard.add('uid')
            uid_obj.value = ''.join([random.choice(choice) for _ in range(36)])
            self.vcard_full_filename = os.path.join(addressbook_path,
                    self.vcard.uid.value + ".vcf")
        else:
            # create vcard from file
            self.vcard_full_filename = filename
            # open .vcf file
            try:
                file = open(filename, "r")
                contents = file.read()
                file.close()
            except IOError as e:
                raise CarddavObject.VCardParseError(e)
            # create vcard object
            try:
                self.vcard = vobject.readOne(contents)
            except vobject.base.ParseError as e:
                # if creation fails, try to repair vcard contents
                try:
                    self.vcard = vobject.readOne(
                            self.filter_invalid_tags(contents))
                    self.write_to_file(overwrite=True)
                except vobject.base.ParseError as e:
                    raise CarddavObject.VCardParseError(e)

    def __str__(self):
        return self.get_full_name()

    def filter_invalid_tags(self, contents):
        contents = re.sub('(?i)' + re.escape('X-messaging/aim-All'), 'X-AIM', contents)
        contents = re.sub('(?i)' + re.escape('X-messaging/gadu-All'), 'X-GADUGADU', contents)
        contents = re.sub('(?i)' + re.escape('X-messaging/groupwise-All'), 'X-GROUPWISE', contents)
        contents = re.sub('(?i)' + re.escape('X-messaging/icq-All'), 'X-ICQ', contents)
        contents = re.sub('(?i)' + re.escape('X-messaging/xmpp-All'), 'X-JABBER', contents)
        contents = re.sub('(?i)' + re.escape('X-messaging/msn-All'), 'X-MSN', contents)
        contents = re.sub('(?i)' + re.escape('X-messaging/yahoo-All'), 'X-YAHOO', contents)
        contents = re.sub('(?i)' + re.escape('X-messaging/skype-All'), 'X-SKYPE', contents)
        contents = re.sub('(?i)' + re.escape('X-messaging/irc-All'), 'X-IRC', contents)
        contents = re.sub('(?i)' + re.escape('X-messaging/sip-All'), 'X-SIP', contents)
        return contents

    def process_user_input(self, input):
        # parse user input string
        contact_data = {}
        counter = 1
        for line in input.splitlines():
            if line == "" or line.startswith("#"):
                continue
            try:
                key = line.split("=")[0].strip().lower()
                value = line.split("=")[1].strip()
                if value == "":
                    continue
                if contact_data.has_key(key):
                    print "Error in input line %d: key %s already exists" % (counter, key)
                    sys.exit(1)
                contact_data[key] = value.decode("utf-8")
                counter += 1
            except IndexError as e:
                print "Error in input line %d: Malformed input\nLine: %s" % (counter, line)
                sys.exit(1)

        # clean vcard
        self.clean_vcard()
        # process data
        # update ref
        dt = datetime.datetime.now()
        rev_obj = self.vcard.add('rev')
        rev_obj.value = "%.4d%.2d%.2dT%.2d%.2d%.2dZ" \
                % (dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second) 

        # name and organisation
        # either enter first name, first and last name or organisation
        if contact_data.has_key("first name") == False:
            contact_data['first name'] = ""
        if contact_data.has_key("last name") == False:
            contact_data['last name'] = ""
        if contact_data.has_key("organisation") == False:
            contact_data['organisation'] = ""
        if (contact_data['first name'] == "" and contact_data['last name'] == "" and contact_data['organisation'] == "") \
                or (contact_data['first name'] == "" and contact_data['last name'] != ""):
            print "Error: You must enter a first name, a first and last name or an organisation"
            sys.exit(1)
        else:
            self.set_name_and_organisation(contact_data['first name'],
                    contact_data['last name'], contact_data['organisation'])

        # phone
        for key in contact_data.keys():
            if key.startswith("phone") == False:
                continue
            try:
                label = contact_data[key].split(":")[0].strip()
                if label == "":
                    print "Error: Missing label for line %s" % key
                    sys.exit(1)
                number = contact_data[key].split(":")[1].strip()
                if number == "":
                    print "Info: Skipped %s" % key
                    continue
            except IndexError as e:
                print "Error: The %s line is malformed" % key
                sys.exit(1)
            self.add_phone_number(label, number)

        # email
        for key in contact_data.keys():
            if key.startswith("email") == False:
                continue
            try:
                label = contact_data[key].split(":")[0].strip()
                if label == "":
                    print "Error: Missing label for line %s" % key
                    sys.exit(1)
                email = contact_data[key].split(":")[1].strip()
                if email == "":
                    print "Info: Skipped %s" % key
                    continue
            except IndexError as e:
                print "Error: The %s line is malformed" % key
                sys.exit(1)
            self.add_email_address(label, email)

        # post addresses
        for key in contact_data.keys():
            if key.startswith("address") == False:
                continue
            try:
                label = contact_data[key].split(":")[0].strip()
                if label == "":
                    print "Error: Missing label for line %s" % key
                    sys.exit(1)
                address = contact_data[key].split(":")[1].strip()
                if address.startswith("; ; ; ; "):
                    print "Info: Skipped %s" % key
                    continue
            except IndexError as e:
                print "Error: The %s line is malformed" % key
                sys.exit(1)
            if address.split(";").__len__() != 5:
                print "Error: The %s line is malformed" % key
                sys.exit(1)
            street_and_house_number = address.split(";")[0].strip()
            if street_and_house_number == "":
                print "Error: %s has no street" % key
                sys.exit(1)
            postcode = address.split(";")[1].strip()
            if postcode == "":
                print "Error: %s has no postcode" % key
                sys.exit(1)
            city = address.split(";")[2].strip()
            if city == "":
                print "Error: %s has no city" % key
                sys.exit(1)
            region = address.split(";")[3].strip()
            country = address.split(";")[4].strip()
            if country == "":
                print "Error: %s has no country" % key
                sys.exit(1)
            self.add_post_address(label, street_and_house_number, postcode, city, region, country)

        # instant messaging and social networks
        if contact_data.has_key("jabber") and contact_data['jabber'] != "":
            self.set_jabber_id(contact_data['jabber'])
        if contact_data.has_key("skype") and contact_data['skype'] != "":
            self.set_skype_id(contact_data['skype'])
        if contact_data.has_key("twitter") and contact_data['twitter'] != "":
            self.set_twitter_id(contact_data['twitter'])
        if contact_data.has_key("webpage") and contact_data['webpage'] != "":
            self.set_webpage(contact_data['webpage'])

        # miscellaneous stuff
        # nickname
        if contact_data.has_key("nickname") and contact_data['nickname'] != "":
            self.set_nickname(contact_data['nickname'])
        # birthday
        if contact_data.has_key("birthday") and contact_data['birthday'] != "":
            try:
                date = datetime.datetime.strptime(contact_data['birthday'], "%d.%m.%Y")
                self.set_birthday(date)
            except ValueError as e:
                print "Error: Birthday date in the wrong format. Example: 31.12.1989"
                sys.exit(1)

    def get_addressbook_name(self):
        return self.addressbook_name

    def get_vcard_full_filename(self):
        return self.vcard_full_filename

    def get_first_name(self):
        try:
            return self.vcard.n.value.given.encode("utf-8")
        except AttributeError as e:
            return ""

    def get_last_name(self):
        try:
            return self.vcard.n.value.family.encode("utf-8")
        except AttributeError as e:
            return ""

    def get_full_name(self):
        try:
            return self.vcard.fn.value.encode("utf-8")
        except AttributeError as e:
            if self.get_first_name() != "" or self.get_last_name() != "":
                return "%s %s" % (self.get_first_name(), self.get_last_name())
            elif self.get_organisation() != "":
                return self.get_organisation()
            else:
                return ""

    def get_categories(self):
        category_list = []
        for child in self.vcard.getChildren():
            if child.name == "CATEGORIES":
                category_list = [child.value[ii].encode('utf-8') for ii in xrange(len(child.value))]
        return category_list

    def get_nickname(self):
        try:
            return self.vcard.nickname.value.encode("utf-8")
        except AttributeError as e:
            return ""

    def set_nickname(self, name):
        nickname_obj = self.vcard.add('nickname')
        nickname_obj.value = name

    def get_organisation(self):
        try:
            return ' '.join(self.vcard.org.value).encode("utf-8")
        except AttributeError as e:
            return ""

    def set_name_and_organisation(self, first_name, last_name, organisation):
        if first_name == "" and last_name == "":
            name_obj = self.vcard.add('fn')
            name_obj.value = organisation
            name_obj = self.vcard.add('n')
            name_obj.value = vobject.vcard.Name(family="", given="")
            showas_obj = self.vcard.add('x-abshowas')
            showas_obj.value = "COMPANY"
        else:
            name_obj = self.vcard.add('fn')
            name_obj.value = "%s %s" % (first_name, last_name)
            name_obj = self.vcard.add('n')
            name_obj.value = vobject.vcard.Name(family=last_name, given=first_name)
        if organisation != "":
            org_obj = self.vcard.add('org')
            org_obj.value = [organisation]

    def get_phone_numbers(self):
        phone_list = []
        for child in self.vcard.getChildren():
            if child.name != "TEL":
                continue
            type = "voice"
            if child.params.has_key("TYPE"):
                type = ','.join(child.params['TYPE']).encode("utf-8")
            elif child.group != None:
                for label in self.vcard.getChildren():
                    if label.name == "X-ABLABEL" and label.group == child.group:
                        type = label.value.encode("utf-8")
                        break
            phone_list.append({"type":type, "value":child.value.encode("utf-8")})
        return phone_list

    def add_phone_number(self, type, number):
        phone_obj = self.vcard.add('tel')
        phone_obj.value = number
        if type.lower() in ["cell", "home", "work",]:
            phone_obj.type_param = type
        else:
            number_of_custom_phone_number_labels = 0
            for label in self.vcard.getChildren():
                if label.name == "X-ABLABEL" and label.group.startswith("itemtel"):
                    number_of_custom_phone_number_labels += 1
            group_name = "itemtel%d" % (number_of_custom_phone_number_labels+1)
            phone_obj.group = group_name
            label_obj = self.vcard.add('x-ablabel')
            label_obj.group = group_name
            label_obj.value = type

    def get_email_addresses(self):
        email_list = []
        for child in self.vcard.getChildren():
            if child.name != "EMAIL":
                continue
            type = "home"
            if child.params.has_key("TYPE"):
                type = ','.join(child.params['TYPE']).encode("utf-8")
            elif child.group != None:
                for label in self.vcard.getChildren():
                    if label.name == "X-ABLABEL" and label.group == child.group:
                        type = label.value.encode("utf-8")
                        break
            email_list.append({"type":type, "value":child.value.encode("utf-8")})
        return email_list

    def add_email_address(self, type, address):
        email_obj = self.vcard.add('email')
        email_obj.value = address
        if type.lower() in ["home", "work",]:
            email_obj.type_param = type
        else:
            number_of_custom_email_labels = 0
            for label in self.vcard.getChildren():
                if label.name == "X-ABLABEL" and label.group.startswith("itememail"):
                    number_of_custom_email_labels += 1
            group_name = "itememail%d" % (number_of_custom_email_labels+1)
            email_obj.group = group_name
            label_obj = self.vcard.add('x-ablabel')
            label_obj.group = group_name
            label_obj.value = type

    def get_post_addresses(self):
        address_list = []
        for child in self.vcard.getChildren():
            if child.name != "ADR":
                continue
            type = "home"
            if child.params.has_key("TYPE"):
                type = ','.join(child.params['TYPE']).encode("utf-8")
            elif child.group != None:
                for label in self.vcard.getChildren():
                    if label.name == "X-ABLABEL" and label.group == child.group:
                        type = label.value.encode("utf-8")
                        break
            try:
                street_and_house_number = child.value.street.encode("utf-8")
            except AttributeError as e:
                street_and_house_number = ', '.join([ x.strip() for x in child.value.street ]).encode("utf-8")
            address_list.append({"type":type,
                    "street_and_house_number":street_and_house_number,
                    "postcode":child.value.code.encode("utf-8"),
                    "city":child.value.city.encode("utf-8"),
                    "region":child.value.region.encode("utf-8"),
                    "country":child.value.country.encode("utf-8")})
        return address_list

    def add_post_address(self, type, street_and_house_number, postcode, city, region, country):
        adr_obj = self.vcard.add('adr')
        adr_obj.value = vobject.vcard.Address(street=street_and_house_number,
                city=city, region=region, code=postcode, country=country)
        if type.lower() in ["home", "work",]:
            adr_obj.type_param = type
        else:
            number_of_custom_post_address_labels = 0
            for label in self.vcard.getChildren():
                if label.name == "X-ABLABEL" and label.group.startswith("itemadr"):
                    number_of_custom_post_address_labels += 1
            group_name = "itemadr%d" % (number_of_custom_post_address_labels+1)
            adr_obj.group = group_name
            label_obj = self.vcard.add('x-ablabel')
            label_obj.group = group_name
            label_obj.value = type

    def get_jabber_id(self):
        try:
            return self.vcard.x_jabber.value.encode("utf-8")
        except AttributeError as e:
            return ""

    def set_jabber_id(self, jabber_id):
        jabber_obj = self.vcard.add('x-jabber')
        jabber_obj.value = jabber_id

    def get_skype_id(self):
        try:
            return self.vcard.x_skype.value.encode("utf-8")
        except AttributeError as e:
            return ""

    def set_skype_id(self, skype_id):
        skype_obj = self.vcard.add('x-skype')
        skype_obj.value = skype_id

    def get_twitter_id(self):
        try:
            return self.vcard.x_twitter.value.encode("utf-8")
        except AttributeError as e:
            return ""

    def set_twitter_id(self, twitter_id):
        twitter_obj = self.vcard.add('x-twitter')
        twitter_obj.value = twitter_id

    def get_webpage(self):
        try:
            return self.vcard.url.value.encode("utf-8")
        except AttributeError as e:
            return ""

    def set_webpage(self, webpage):
        webpage_obj = self.vcard.add('url')
        webpage_obj.value = webpage

    def get_birthday(self):
        """:returns: contacts birthday or None if not available
            :rtype: datetime.datetime
        """
        try:
            return datetime.datetime.strptime(self.vcard.bday.value.replace('-', ''), "%Y%m%d")
        except AttributeError as e:
            return None
        except ValueError as e:
            return None

    def set_birthday(self, date):
        bday_obj = self.vcard.add('bday')
        bday_obj.value = "%.4d%.2d%.2d" % (date.year, date.month, date.day)

    def print_vcard(self):
        strings = ["Name: %s" % self.get_full_name()]
        if self.get_categories().__len__() > 0:
            strings.append("Categories\n    %s" % ', '.join(self.get_categories()))
        if self.get_organisation() != "" \
                and self.get_organisation() != self.get_full_name():
            strings.append("organisation: %s" % self.get_organisation())
        if self.get_nickname() != "":
            strings.append("Nickname: %s" % self.get_nickname())
        if self.get_phone_numbers().__len__() > 0:
            strings.append("Phone")
            for index, entry in enumerate(self.get_phone_numbers()):
                strings.append("    %s: %s" % (entry['type'], entry['value']))
        if self.get_email_addresses().__len__() > 0:
            strings.append("E-Mail")
            for index, entry in enumerate(self.get_email_addresses()):
                strings.append("    %s: %s" % (entry['type'], entry['value']))
        if self.get_post_addresses().__len__() > 0:
            strings.append("Addresses")
            for index, entry in enumerate(self.get_post_addresses()):
                strings.append("    %s:" % entry['type'])
                if entry['street_and_house_number'] != "":
                    strings.append("        %s" % entry['street_and_house_number'])
                if entry['postcode'] != "" and entry['city'] != "":
                    strings.append("        %s %s" % (entry['postcode'], entry['city']))
                elif entry['postcode'] != "":
                    strings.append("        %s" % entry['postcode'])
                elif entry['city'] != "":
                    strings.append("        %s" % entry['city'])
                if entry['region'] != "" and entry['country'] != "":
                    strings.append("        %s, %s" % (entry['region'], entry['country']))
                elif entry['region'] != "":
                    strings.append("        %s" % entry['region'])
                elif entry['country'] != "":
                    strings.append("        %s" % entry['country'])
        if self.get_jabber_id() != "" \
                or self.get_skype_id() != "" \
                or self.get_twitter_id() != "" \
                or self.get_webpage() != "":
            strings.append("Instant messaging and social networks")
            if self.get_jabber_id() != "":
                strings.append("    Jabber:  %s" % self.get_jabber_id())
            if self.get_skype_id() != "":
                strings.append("    Skype:   %s" % self.get_skype_id())
            if self.get_twitter_id() != "":
                strings.append("    Twitter: %s" % self.get_twitter_id())
            if self.get_webpage() != "":
                strings.append("    Webpage: %s" % self.get_webpage())
        if self.get_birthday() != None:
            strings.append("Miscellaneous")
            if self.get_birthday() != None:
                date = self.get_birthday()
                strings.append("    Birthday: %.2d.%.2d.%.4d" % (date.day, date.month, date.year))
        return '\n'.join(strings)

    def write_to_file(self, overwrite=False):
        if os.path.exists(self.vcard_full_filename) and overwrite == False:
            print "Error: vcard with the file name %s already exists" \
                    % os.path.basename(self.vcard_full_filename)
            sys.exit(4)
        try:
            vcard_output = self.vcard.serialize()
            file = open(self.vcard_full_filename, "w")
            file.write(vcard_output)
            file.close()
        except vobject.base.ValidateError as e:
            print "Error: Vcard is not valid.\n%s" % e
            sys.exit(4)
        except IOError as e:
            print "Error: Can't write\n%s" % e
            sys.exit(4)

    def delete_vcard_file(self):
        if os.path.exists(self.vcard_full_filename):
            os.remove(self.vcard_full_filename)
        else:
            print "Error: Vcard file %s does not exist." % self.vcard_full_filename
            sys.exit(4)

    def clean_vcard(self):
        # rev
        try:
            self.vcard.remove(self.vcard.rev)
        except AttributeError as e:
            pass
        # name
        try:
            self.vcard.remove(self.vcard.n)
            self.vcard.remove(self.vcard.fn)
        except AttributeError as e:
            pass
        # organisation
        try:
            self.vcard.remove(self.vcard.org)
        except AttributeError as e:
            pass
        try:
            self.vcard.remove(self.vcard.x_abshowas)
        except AttributeError as e:
            pass
        # phone
        while True:
            try:
                self.vcard.remove(self.vcard.tel)
            except AttributeError as e:
                break
        # email addresses
        while True:
            try:
                self.vcard.remove(self.vcard.email)
            except AttributeError as e:
                break
        # addresses
        while True:
            try:
                self.vcard.remove(self.vcard.adr)
            except AttributeError as e:
                break
        # instant messaging and social networks
        try:
            self.vcard.remove(self.vcard.x_jabber)
        except AttributeError as e:
            pass
        try:
            self.vcard.remove(self.vcard.x_skype)
        except AttributeError as e:
            pass
        try:
            self.vcard.remove(self.vcard.x_twitter)
        except AttributeError as e:
            pass
        try:
            self.vcard.remove(self.vcard.url)
        except AttributeError as e:
            pass
        # nickname
        try:
            self.vcard.remove(self.vcard.nickname)
        except AttributeError as e:
            pass
        # birthday
        try:
            self.vcard.remove(self.vcard.bday)
        except AttributeError as e:
            pass
        # x-ablabel
        while True:
            try:
                self.vcard.remove(self.vcard.x_ablabel)
            except AttributeError as e:
                break

    class VCardParseError(LookupError):
        """ is called, when vcard could not be parsed """
